#include "siftop.h"
#include "galias.h"
//#include "siftop_element_types.h"
//#include "cmlfile.h"
#include <stdlib.h>
#include <string.h>

const gtypel operator ++(gtypel &a, int)
{
  gtypel old;
  switch (a)
  {
    case isolinear1d:
      old = a;
      a = isoquadratic1d;
      return old;
    case isoquadratic1d:
      old = a;
      a = trianglelinear;
      return old;
    case trianglelinear:
      old = a;
      a = trianglequadratic;
      return old;
    case trianglequadratic:
      old = a;
      a = isolinear2d;
      return old;
    case isolinear2d:
      old = a;
      a = isoquadratic2d;
      return old;
    case isoquadratic2d:
      old = a;
      a = tetrahedronlinear;
      return old;
    case tetrahedronlinear:
      old = a;
      a = tetrahedronquadratic;
      return old;
    case tetrahedronquadratic:
      old = a;
      a = pyramidelinear;
      return old;
    case pyramidelinear:
      old = a;
      a = pyramidequadratic;
      return old;
    case pyramidequadratic:
      old = a;
      a = vedgelinear;
      return old;
    case vedgelinear:
      old = a;
      a = vedgequadratic;
      return old;
    case vedgequadratic:
      old = a;
      a = isolinear3d;
      return old;
    case isolinear3d:
      old = a;
      a = isoquadratic3d;
      return old;
    case isoquadratic3d:
      old = a;
      a = all_element_types;
      return old;
    default:
      print_err("increasing of invalid gtypel value is required", __FILE__, __LINE__, __func__);
  }
  return all_element_types;
}



/**
  Constructor setups all data members to zero values
  
  Created by Tomas Koudelka 03.2008, koudelka@cml.fsv.cvut.cz
*/
snode::snode()
{
  x = y = z = 0.0;
  nprop = 0;
  entid = NULL;
  prop  = NULL;
}



/**
  Destructor deletes dynamically allocated data members
  
  Created by Tomas Koudelka 03.2008, koudelka@cml.fsv.cvut.cz
*/
snode::~snode()
{
  delete [] entid;
  delete [] prop;
}


/**
  This method allocates array properties which size is given by num_prop

  @param num_prop - number of nodes of given element i.e. size of node array

  Created by Tomas Koudelka 03.2008, koudelka@cml.fsv.cvut.cz
*/
void snode::alloc(long num_prop)
{
  nprop = num_prop;
  entid = new entityp[nprop];
  prop  = new long[nprop];
}



/**
  This method searches array of properties for property aprop with given entity type ent.
  If the pointers edg and sfc are not NULL then the additional edges edg or surfaces sfc are
  also searched for the given property.

  @param prop - searched property number 
  @param ent  - general entity type whose property is searched
  @param nid  - node id (it is used in connection with the additional edg and sfc data)
  @param edg  - pointer to structure with data about additional edges
  @param sfc  - pointer to structure with data about additional surfaces

  @retval 0 - searched property was not found on given entity type
  @retval 1 - searched property was found

  Created by Tomas Koudelka 03.2008, koudelka@cml.fsv.cvut.cz
*/
long snode::searchprop(long aprop, gentity ent, long nid, sedges *edg, sfaces *sfc)
{
  long i, j;
  for(i=0; i<nprop; i++)
  {
    switch(ent)
    {
      case gvertex:
      case gcurve:
      case gregion: 
        if ((int(ent) == int(entid[i])) && (prop[i] == aprop))   return 1;
        break;
      case gsurface:
        if (((entid[i] == esurface) || (entid[i] == epatch) || 
             (entid[i] == eshell)) && (prop[i] == aprop))    return 1;
	break;
      default:
        print_err("Unknown type of entity is required", __FILE__, __LINE__, __func__);
        abort();
    }
  }
 
  if ((ent == gcurve) && (edg))
  {
    // search additional edges specified in the MeshEditor
    for(i=0; i<edg->nprop; i++)
    {
      if (edg->proplist[i] == aprop)
      {
        for(j=0; j<edg->prop_list_nod_num[i]; j++)
        {
          if (edg->prop_list_nod[i][j] == nid)   return 1;
          if (edg->prop_list_nod[i][j] > nid)    break;  // node numbers are sorted ascendently in the prop_list_nod
        }
      }
    }
  }

  if ((ent == gsurface) && (sfc))
  {
    // search additional surfacs specified in the MeshEditor
    for(i=0; i<sfc->nprop; i++)
    {
      if (sfc->proplist[i] == aprop)
      {
        for(j=0; j<sfc->prop_list_nod_num[i]; j++)
        {
          if (sfc->prop_list_nod[i][j] == nid)   return 1;
          if (sfc->prop_list_nod[i][j] > nid)    break;  // node numbers are sorted ascendently in the prop_list_nod
        }
      }
    }
  }
  return 0;
}



/**
  Constructor setups all data members to zero values
  
  Created by Tomas Koudelka 03.2008, koudelka@cml.fsv.cvut.cz
*/
selement::selement()
{
  type = gtypel(0);
  nne = 0L;

  ned =  tnned = 0L;
  nned     = NULL;
  propedg  = NULL;
  edgenod  = NULL;

  nsurf = tnnsurf = 0;
  nnsurf   = NULL;
  surfnod  = NULL;
  propsurf = NULL;  

  nodes    = NULL;
  prop     = 0L;
}



/**
  Destructor deletes dynamically allocated data members
  
  Created by Tomas Koudelka 03.2008, koudelka@cml.fsv.cvut.cz
*/
selement::~selement()
{
  long i;
  delete [] nned;
  delete [] nnsurf;
  delete [] nodes;
  delete [] propedg;
  delete [] propsurf;
  for(i=0; i<ned; i++)
    delete [] edgenod[i];
  for(i=0; i<nsurf; i++)
    delete [] surfnod[i];
  delete [] edgenod;
  delete [] surfnod;
}



/**
  This method allocates array nodes depending on the element type.
  Element type is given by data member type.

  @param alloc_bp - controls allocation of boundary properties arrays on elements 
                    (1 = wil be allocated/ 0 = will not be allocated)

  @retval 0 : on success allocation
  @retval 1 : unknown type is required


  Created by Tomas Koudelka 03.2008, koudelka@cml.fsv.cvut.cz
*/
long selement::alloc(long alloc_bp)
{
  long i;
  switch (type)
  {
    case isolinear1d:
      nne   = 2;
      ned   = 1;
      nned  = new long[ned];
      edgenod = new long*[ned];
      for(i=0; i<ned; i++)
      {
        nned[i] = 2;
        edgenod[i] = new long[nned[i]];
      }
      edgenod[0][0] = 0;
      edgenod[0][1] = 1;
      nsurf = 0;
      tnned = 2;
      tnnsurf = 0;
      break;
    case isoquadratic1d:
      nne   = 3;
      ned   = 1;
      nned  = new long[ned];
      edgenod = new long*[ned];
      for(i=0; i<ned; i++)
      {
        nned[i] = 3;
        edgenod[i] = new long[nned[i]];
      }
      edgenod[0][0] = 0;
      edgenod[0][1] = 1;
      edgenod[0][2] = 2;
      nsurf = 0;
      tnned = 3;
      tnnsurf = 0;
      break;
    case trianglelinear:
      nne   = 3;
      ned   = 3;
      nsurf = 1;
      nned  = new long[ned];
      edgenod = new long*[ned];
      surfnod = new long*[nsurf];
      for(i=0; i<ned; i++)
      {
        nned[i] = 2;
        edgenod[i] = new long[nned[i]];        
      }
      edgenod[0][0] = 0;  edgenod[0][1] = 1;
      edgenod[1][0] = 1;  edgenod[1][1] = 2;
      edgenod[2][0] = 2;  edgenod[2][1] = 0;
      nnsurf = new long[1];
      for(i=0; i<nsurf; i++)
      {
        nnsurf[i] = 3;
        surfnod[i] = new long[nnsurf[i]];        
      }
      surfnod[0][0] = 0;  surfnod[0][1] = 1;  surfnod[0][2] = 2;
      tnned = 6;
      tnnsurf = 3; 
      break;
    case trianglequadratic:
      nne   = 6;
      ned   = 3;
      nsurf = 1;
      nned  = new long[ned];
      edgenod = new long*[ned];
      surfnod = new long*[nsurf];
      for(i=0; i<ned; i++)
      {
        nned[i] = 3;
        edgenod[i] = new long[nned[i]];        
      }
      edgenod[0][0] = 0;  edgenod[0][1] = 1;  edgenod[0][2] = 3;
      edgenod[1][0] = 1;  edgenod[1][1] = 2;  edgenod[1][2] = 4;
      edgenod[2][0] = 2;  edgenod[2][1] = 0;  edgenod[2][2] = 5;
      nnsurf = new long[nsurf];
      for(i=0; i<nsurf; i++)
      {
        nnsurf[i] = 6;
        surfnod[i] = new long[nnsurf[i]];        
      }
      surfnod[0][0] = 0;  surfnod[0][1] = 1;  surfnod[0][2] = 2;
      surfnod[0][3] = 3;  surfnod[0][4] = 4;  surfnod[0][5] = 5;
      tnned = 9;
      tnnsurf = 6;
      break;
    case isolinear2d:
      nne   = 4;
      ned   = 4;
      nsurf = 1;
      nned  = new long[ned];
      edgenod = new long*[ned];
      surfnod = new long*[nsurf];
      for(i=0; i<ned; i++)
      {
        nned[i] = 2;
        edgenod[i] = new long[nned[i]];        
      }
      edgenod[0][0] = 0;  edgenod[0][1] = 1;
      edgenod[1][0] = 1;  edgenod[1][1] = 2;
      edgenod[2][0] = 2;  edgenod[2][1] = 3;
      edgenod[3][0] = 3;  edgenod[3][1] = 0;
      nnsurf = new long[nsurf];
      for(i=0; i<nsurf; i++)
      {
        nnsurf[i] = 4;
        surfnod[i] = new long[nnsurf[i]];        
      }
      surfnod[0][0] = 0;  surfnod[0][1] = 1;  surfnod[0][2] = 2;  surfnod[0][3] = 3;
      tnned = 8;
      tnnsurf = 4;
      break;
    case isoquadratic2d:
      nne   = 8;
      ned   = 4;
      nsurf = 1;
      nned  = new long[ned];
      edgenod = new long*[ned];
      surfnod = new long*[nsurf];
      for(i=0; i<ned; i++)
      {
        nned[i] = 3;
        edgenod[i] = new long[nned[i]];        
      }
      edgenod[0][0] = 0;  edgenod[0][1] = 1;  edgenod[0][2] = 4;
      edgenod[1][0] = 1;  edgenod[1][1] = 2;  edgenod[1][2] = 5;
      edgenod[2][0] = 2;  edgenod[2][1] = 3;  edgenod[2][2] = 6;
      edgenod[3][0] = 3;  edgenod[3][1] = 0;  edgenod[3][2] = 7;
      nnsurf = new long[nsurf];
      for(i=0; i<nsurf; i++)
      {
        nnsurf[i] = 8;
        surfnod[i] = new long[nnsurf[i]];        
      }
      surfnod[0][0] = 0;  surfnod[0][1] = 1;  surfnod[0][2] = 2;  surfnod[0][3] = 3;
      surfnod[0][4] = 4;  surfnod[0][5] = 5;  surfnod[0][6] = 6;  surfnod[0][7] = 7;
      tnned = 12;
      tnnsurf = 8;
      break;
    case tetrahedronlinear:
      nne   = 4;
      ned   = 6;
      nsurf = 4;
      nned  = new long[ned];
      edgenod = new long*[ned];
      surfnod = new long*[nsurf];
      for(i=0; i<ned; i++)
      {
        nned[i] = 2;
        edgenod[i] = new long[nned[i]];        
      }
      edgenod[0][0]=0;  edgenod[0][1]=1;
      edgenod[1][0]=1;  edgenod[1][1]=2; 
      edgenod[2][0]=2;  edgenod[2][1]=0; 
      edgenod[3][0]=1;  edgenod[3][1]=3; 
      edgenod[4][0]=3;  edgenod[4][1]=2; 
      edgenod[5][0]=3;  edgenod[5][1]=0; 
      nnsurf = new long[nsurf];
      for(i=0; i<nsurf; i++)
      {
        nnsurf[i] = 3;
        surfnod[i] = new long[nnsurf[i]];        
      }
      surfnod[0][0]=2;  surfnod[0][1]=1; surfnod[0][2]=3; 
      surfnod[1][0]=0;  surfnod[1][1]=2; surfnod[1][2]=3; 
      surfnod[2][0]=1;  surfnod[2][1]=0; surfnod[2][2]=3; 
      surfnod[3][0]=0;  surfnod[3][1]=1; surfnod[3][2]=2; 
      tnned = 12;
      tnnsurf = 12;
      break;
    case tetrahedronquadratic:
      nne   = 10;
      ned   = 6;
      nsurf = 4;
      nned  = new long[ned];
      edgenod = new long*[ned];
      surfnod = new long*[nsurf];
      for(i=0; i<ned; i++)
      {
        nned[i] = 3;
        edgenod[i] = new long[nned[i]];        
      }
      edgenod[0][0]=0;  edgenod[0][1]=1;  edgenod[0][2]=4;
      edgenod[1][0]=1;  edgenod[1][1]=2;  edgenod[1][2]=5;
      edgenod[2][0]=2;  edgenod[2][1]=0;  edgenod[2][2]=6;
      edgenod[3][0]=1;  edgenod[3][1]=3;  edgenod[3][2]=8;
      edgenod[4][0]=3;  edgenod[4][1]=2;  edgenod[4][2]=9;
      edgenod[5][0]=3;  edgenod[5][1]=0;  edgenod[5][2]=7;
      nnsurf = new long[nsurf];
      for(i=0; i<nsurf; i++)
      {
        nnsurf[i] = 6;
        surfnod[i] = new long[nnsurf[i]];        
      }
      surfnod[0][0]=2;  surfnod[0][1]=1; surfnod[0][2]=3;
      surfnod[0][3]=5;  surfnod[0][4]=8; surfnod[0][5]=9;  
      surfnod[1][0]=0;  surfnod[1][1]=2; surfnod[1][2]=3;
      surfnod[1][3]=6;  surfnod[1][4]=9; surfnod[1][5]=7;  
      surfnod[2][0]=1;  surfnod[2][1]=0; surfnod[2][2]=3; 
      surfnod[2][3]=4;  surfnod[2][4]=7; surfnod[2][5]=8;   
      surfnod[3][0]=0;  surfnod[3][1]=1; surfnod[3][2]=2; 
      surfnod[3][3]=4;  surfnod[3][4]=5; surfnod[3][5]=6;  
      tnned = 18;
      tnnsurf = 24;
      break;
    case pyramidelinear:
      nne   = 5;
      ned   = 8;
      nsurf = 5;
      nned  = new long[ned];
      for(i=0; i<ned; i++)
      {
        nned[i] = 2;
        edgenod[i] = new long[nned[i]];        
      }
      nnsurf = new long[nsurf];
      for(i=0; i<nsurf-1; i++)
      {
        nnsurf[i] = 3;
        surfnod[i] = new long[nnsurf[i]];        
      }
      nnsurf[nsurf-1] = 4;
      surfnod[nsurf-1] = new long[nnsurf[nsurf-1]];
      tnned = 16;
      tnnsurf = 16;
      break;
    case pyramidequadratic:
      nne   = 13;
      ned   = 9;
      nsurf = 5;
      for(i=0; i<ned; i++)
      {
        nned[i] = 3;
        edgenod[i] = new long[nned[i]];        
      }
      nnsurf = new long[nsurf];
      for(i=0; i<nsurf-1; i++)
      {
        nnsurf[i] = 6;
        surfnod[i] = new long[nnsurf[i]];        
      }
      nnsurf[nsurf-1] = 8;
      surfnod[nsurf-1] = new long[nnsurf[nsurf-1]];
      tnned = 27;
      tnnsurf = 32;
      break;
    case vedgelinear:
      nne   = 6;
      ned   = 9;
      nsurf = 5;
      for(i=0; i<ned; i++)
      {
        nned[i] = 2;
        edgenod[i] = new long[nned[i]];        
      }
      nnsurf = new long[nsurf];
      for(i=0; i<nsurf-2; i++)
      {
        nnsurf[i] = 4;
        surfnod[i] = new long[nnsurf[i]];        
      }
      nnsurf[nsurf-2] = 3;
      nnsurf[nsurf-1] = 3;
      surfnod[nsurf-2] = new long[nnsurf[nsurf-2]];
      surfnod[nsurf-1] = new long[nnsurf[nsurf-1]];
      tnned = 18;
      tnnsurf = 18;
      break;
    case vedgequadratic:
      nne   = 15;
      ned   = 9;
      nsurf = 5;
      for(i=0; i<ned; i++)
      {
        nned[i] = 3;
        edgenod[i] = new long[nned[i]];        
      }
      nnsurf = new long[nsurf];
      for(i=0; i<nsurf-2; i++)
      {
        nnsurf[i] = 8;
        surfnod[i] = new long[nnsurf[i]];        
      }
      nnsurf[nsurf-2] = 6;
      nnsurf[nsurf-1] = 6;
      surfnod[nsurf-2] = new long[nnsurf[nsurf-2]];
      surfnod[nsurf-1] = new long[nnsurf[nsurf-1]];
      tnned = 27;
      tnnsurf = 36;
      break;
    case isolinear3d:
      nne   = 8;
      ned   = 12;
      nsurf = 6;
      nned  = new long[ned];
      edgenod = new long*[ned];
      surfnod = new long*[nsurf];
      for(i=0; i<ned; i++)
      {
        nned[i] = 2;
        edgenod[i] = new long[nned[i]];        
      }
      edgenod[0][0] =0;  edgenod[0][1] =1; 
      edgenod[1][0] =1;  edgenod[1][1] =2; 
      edgenod[2][0] =2;  edgenod[2][1] =3; 
      edgenod[3][0] =3;  edgenod[3][1] =0; 
      edgenod[4][0] =0;  edgenod[4][1] =4; 
      edgenod[5][0] =1;  edgenod[5][1] =5; 
      edgenod[6][0] =2;  edgenod[6][1] =6; 
      edgenod[7][0] =3;  edgenod[7][1] =7; 
      edgenod[8][0] =4;  edgenod[8][1] =5; 
      edgenod[9][0] =5;  edgenod[9][1] =6; 
      edgenod[10][0]=6;  edgenod[10][1]=7; 
      edgenod[11][0]=7;  edgenod[11][1]=4; 
      nnsurf = new long[nsurf];
      for(i=0; i<nsurf; i++)
      {
        nnsurf[i] = 4;
        surfnod[i] = new long[nnsurf[i]];        
        linhexahedral_surfnod (surfnod[i], i);
      }
      tnned = 24;
      tnnsurf = 24;
      break;
    case isoquadratic3d:
      nne   = 20;
      ned   = 12;
      nsurf = 6;
      edgenod = new long*[ned];
      surfnod = new long*[nsurf];
      for(i=0; i<ned; i++)
      {
        nned[i] = 3;
        edgenod[i] = new long[nned[i]];        
      }
      edgenod[0][0] =0;  edgenod[0][1] =1;  edgenod[0][2] = 8; 
      edgenod[1][0] =1;  edgenod[1][1] =2;  edgenod[1][2] = 9;
      edgenod[2][0] =2;  edgenod[2][1] =3;  edgenod[2][2] = 10;
      edgenod[3][0] =3;  edgenod[3][1] =0;  edgenod[3][2] = 11;
      edgenod[4][0] =0;  edgenod[4][1] =4;  edgenod[4][2] = 12;
      edgenod[5][0] =1;  edgenod[5][1] =5;  edgenod[5][2] = 13;
      edgenod[6][0] =2;  edgenod[6][1] =6;  edgenod[6][2] = 14;
      edgenod[7][0] =3;  edgenod[7][1] =7;  edgenod[7][2] = 15;
      edgenod[8][0] =4;  edgenod[8][1] =5;  edgenod[8][2] = 16;
      edgenod[9][0] =5;  edgenod[9][1] =6;  edgenod[9][2] = 17;
      edgenod[10][0]=6;  edgenod[10][1]=7;  edgenod[10][2]= 18;
      edgenod[11][0]=7;  edgenod[11][1]=4;  edgenod[11][2]= 19;
      nnsurf = new long[nsurf];
      for(i=0; i<nsurf; i++)
      {
        nnsurf[i] = 8;
        surfnod[i] = new long[nnsurf[i]];        
        quadhexahedral_surfnod (surfnod[i],i);
      }

      tnned = 36;
      tnnsurf = 48;
      break;
    default:
      return 1;
  }
  nodes    = NULL;
  propedg  = NULL;
  propsurf = NULL;
  if (nne)
  {
    nodes = new long[nne];
    memset(nodes, 0, sizeof(*nodes)*nne);
  }
  if (ned && alloc_bp)
  {
    propedg = new long[ned];
    memset(propedg, 0, sizeof(*propedg)*ned);
  }
  if (nsurf && alloc_bp)
  {
    propsurf = new long[nsurf];
    memset(propsurf, 0, sizeof(*propsurf)*nsurf);
  }
  return 0;
}



/**
  This method searches array of properties for property aprop with given entity type ent.
  If the pointers edg and sfc are not NULL then the additional edges edg or surfaces sfc are
  also searched for the given property.

  @param prop - searched property number 
  @param ent  - general entity type whose property is searched
  @param eid  - element id (it is used in connection with the additional edg and sfc data)
  @param edg  - pointer to structure with data about additional edges
  @param sfc  - pointer to structure with data about additional surfaces
  @param entid - pointer to the array of indeces of given entity with given property on the element.
                 It is output parameter used only in connection with additional edges and surfaces
  @param nentid - the length of array entid (output parameter)

  @retval -1 - no property of edges or surfaces was given for the mesh
  @retval  0 - searched property was not found on given entity type
  @retval  1 - searched property was found

  Created by Tomas Koudelka 06.2009, koudelka@cml.fsv.cvut.cz
*/
long selement::searchprop(long aprop, gentity ent, long eid, sedges *edg, sfaces *sfc, long *&entid, long &nentid)
{
  long i, j, ret = 0;
  entid = NULL;
  nentid = 0;

  switch(ent)
  {
    case gcurve:
      if (propedg)
      {
        for (i=0; i<ned; i++)
        {
          if (propedg[i] == aprop)
            return 1; 
        }
      }
      else
        ret=-1;
      break;
    case gsurface:
      if (propsurf)
      {
        for (i=0; i<nsurf; i++)
        {
          if (propsurf[i] == aprop)
            return 1; 
        }
      }
      else
        ret=-1;
      break;
    case gregion: 
      if (prop == aprop)   return 1;
      break;
    default:
     print_err("Unknown type of entity is required", __FILE__, __LINE__, __func__);
     abort();
  }

  if ((ent == gcurve) && (edg))
  {
    ret = 0; // edge properties was read
    // search additional edges specified in the MeshEditor
    for(i=0; i<edg->nprop; i++)
    {
      if (edg->proplist[i] == aprop)
      {
        for(j=0; j<edg->prop_list_elem_num[i]; j++)
        {
          if (edg->prop_list_elem[i][j] == eid)
          {
            entid = edg->prop_list_elem_edgid[i][j];
            nentid = edg->prop_list_elem_edgid_num[i][j]; 
            return 1;
          }
          if (edg->prop_list_elem[i][j] > eid)    break;  // element numbers are sorted ascendently in the prop_list_elem
        }
      }
    }
  }
  
  if ((ent == gsurface) && (sfc))
  {
    ret = 0; // surface properties was read
    // search additional surfaces specified in the MeshEditor
    for(i=0; i<sfc->nprop; i++)
    {
      if (sfc->proplist[i] == aprop)
      {
        for(j=0; j<sfc->prop_list_elem_num[i]; j++)
        {
          if (sfc->prop_list_elem[i][j] == eid)
          {
            entid = sfc->prop_list_elem_sfid[i][j];
            nentid = sfc->prop_list_elem_sfid_num[i][j]; 
            return 1;
          }
          if (sfc->prop_list_elem[i][j] > eid)    break;  // element numbers are sorted ascendently in the prop_list_elem
        }
      }
    }
  }
  return ret;
}



/**
  The method compares array of nodes nod with node numbers on particular element edges.
  If they corresponds with some of the element edge nodes than the edge id is returned.
  Otherwise the -1 is returned.
  

  @param nod - array of searched nodes
  @param n  - number of searched nodes (length of nod)

  @retval  edgid - searched nodes corresponded with edge edgid
  @retval -1 - no corresondence of the nodes was found

  Created by Tomas Koudelka 08.2011, koudelka@cml.fsv.cvut.cz
*/
long selement::compare_edg(long *nod, long n)
{
  long i, j, k;
  long nf; // number of found nodes from the array nod
  long stop;

  for (i=0; i<ned; i++)
  {
    nf = 0; 
    for(j=0; j<n; j++)
    {
      stop = 1;
      for(k=0; k<nned[i]; k++)
      {
        if (nodes[edgenod[i][k]] == nod[j])
        {
          nf++;
          stop = 0;
          break;
        }        
      }
      if (stop)
        break;
    }
    if (nf == n)
      return i;
  }
  return -1;
}



/**
  The method compares array of nodes nod with node numbers on particular element surfaces.
  If they corresponds with some of the element surface nodes than the surface id is returned.
  Otherwise the -1 is returned.
  

  @param nod - array of searched nodes
  @param n  - number of searched nodes (length of nod)

  @retval  surfid - searched nodes corresponded with surface surfid
  @retval -1 - no corresondence of the nodes was found

  Created by Tomas Koudelka 08.2011, koudelka@cml.fsv.cvut.cz
*/
long selement::compare_surf(long *nod, long n)
{
  long i, j, k;
  long nf; // number of found nodes from the array nod
  long stop;

  for (i=0; i<nsurf; i++)
  {
    nf = 0;
    for(j=0; j<n; j++)
    {
      stop = 1;
      for(k=0; k<nnsurf[i]; k++)
      {
        if (nodes[surfnod[i][k]] == nod[j])
        {
          nf++;
          stop = 0;
          break;
        }        
      }
      if (stop)
        break;
    }
    if (nf == n)
      return i;
  }
  return -1;
}



/**
  The constructor initializes all data members to -1.

  Created by Tomas Koudelka, 08.2011
*/
sedge::sedge(void)
{
  n1 = n2 = -1;
  prop = -1;
}



/**
  The destructor specifired only for the formal purposes.

  Created by Tomas Koudelka, 08.2011
*/
sedge::~sedge(void)
{

}



/**
  The function reads data about one edge from the opened text file.

  @param in - pointer to the opened text file

  @retval 0 - on success
  @retval 1 - in the case of an read error

  Created by Tomas Koudelka, 08.2011
*/
long sedge::read(XFILE *in)
{
  if (xfscanf(in, "%ld %ld %ld", &n1, &n2, &prop) < 3)
    return 1;
  // change to C/C++ index notation
  n1--;
  n2--;
  return 0;
}



/**
  The function prints data about one edge to the opened text file.
  If the second parameter lnn is not NULL, 
  then the node numbers on the edge are renumbered using this array.
  The lnn array is used in connection with the mesh decompoistion and it
  contains local node numbers for a domain. The size of the array is given
  by the total number of nodes in the problem.

  @param in - pointer to the opened text file
  @param lnn - pointer to array of local node numbers used for the
               mesh decomposition. Default value of this parameter is NULL.

  @return The function does not return anything.

  Created by Tomas Koudelka, 08.2011
*/
void sedge::print(FILE *out, long *lnn)
{
  if (lnn)
    fprintf(out, "%ld %ld %ld\n", lnn[n1], lnn[n2], prop);
  else
    fprintf(out, "%ld %ld %ld\n", n1+1, n2+1, prop);
}



/**
  The function tests whether all nodes of the edge are contained
  in the given subdomain whose local node numbers given in 
  the parameter lnn. The lnn array is used in connection with
  the mesh decompoistion and it contains local node numbers for a domain. The size 
  of the array is given by the total number of nodes in the problem.

  @param lnn - pointer to array of local node numbers used for the
               mesh decomposition. Default value of this parameter is NULL.

  @retval 0 - edge is not contained in the given subdomain
  @retval 1 - edge is contained in the given subdomain

  Created by Tomas Koudelka, 09.2011
*/
long sedge::cordomtest(long *lnn)
{
  if (lnn[n1] && lnn[n2])
    return 1;

  return 0;
}



/**
  The constructor initializes all data members to zero or NULL values.

  Created by Tomas Koudelka, 08.2011
*/
sedges::sedges()
{
  nedg = 0L;
  nprop = 0L;
  edges = NULL;
  proplist = NULL;
  prop_list_nod = NULL;
  prop_list_nod_num = NULL;
  prop_list_elem = NULL;
  prop_list_elem_num = NULL;
  prop_list_elem_edgid = NULL;
  prop_list_elem_edgid_num = NULL;
}



/**
  The destructor releases all allocated memory.

  Created by Tomas Koudelka, 08.2011
*/
sedges::~sedges()
{
  long i, j;

  delete [] edges;
  for (i=0; i<nprop; i++)
  {
    for (j=0; j<prop_list_elem_num[i]; j++)
      delete [] prop_list_elem_edgid[i][j];
    
    delete [] prop_list_nod[i];
    delete [] prop_list_elem[i];
    delete [] prop_list_elem_edgid[i];
    delete [] prop_list_elem_edgid_num[i];
  }
  if (nprop)
  {
    delete [] proplist;
    delete [] prop_list_nod;
    delete [] prop_list_nod_num;
    delete [] prop_list_elem;
    delete [] prop_list_elem_num;
    delete [] prop_list_elem_edgid;
    delete [] prop_list_elem_edgid_num;
  }
}



/**
  The constructor allocates memory for the list of edges
  and initializes rest of data members to zero or NULL values.
  
  @param n - number of allocated edges

  Created by Tomas Koudelka, 08.2011
*/
sedges::sedges(long n)
{
  nedg = n;
  nprop = 0L;
  edges = new sedge[n];
  proplist = NULL;
  prop_list_nod = NULL;
  prop_list_nod_num = NULL;
  prop_list_elem = NULL;
  prop_list_elem_num = NULL;
  prop_list_elem_edgid = NULL;
  prop_list_elem_edgid_num = NULL;
}



/**
  The function reads list of additional edges from the opened text file.
  The list is placed at the end of file with topology and it is created by
  MeshEditor usually.

  @param in - pointer to the opened text file

  @return The function does not return anything.

  Created by Tomas Koudelka, 08.2011
*/
void sedges::read(XFILE *in)
{
  long i;
  for (i=0; i<nedg; i++)
  {
    if (edges[i].read(in))
    {
      print_err("cannot read edge number %ld", __FILE__, __LINE__, __func__, i+1);
      abort();
    }
  }
}



/**
  The function prints list of additional edges from the opened text file.
  The list is placed the end of file with topology usually in MeshEditor format.
  If the second parameter lnn is not NULL, then only the edges whose all local node numbers
  are nonzero are printed. The lnn array is used in connection with
  the mesh decompoistion and it contains local node numbers for a domain. The size 
  of the array is given by the total number of nodes in the problem.

  @param out - pointer to the opened text file
  @param lnn - pointer to array of local node numbers used for the
               mesh decomposition. Default value of this parameter is NULL.

  @return The function does not return anything.

  Created by Tomas Koudelka, 08.2011
*/
void sedges::print(FILE *out, long *lnn)
{
  long i;
  long *aux, nedg_l;

  if (lnn == NULL) // no mesh decomposition
  {
    fprintf(out, "edges %ld\n", nedg);
    for (i=0; i<nedg; i++)
      edges[i].print(out, lnn);
  }
  else
  {
    nedg_l=0;
    aux = new long[nedg];
    memset(aux, 0, sizeof(*aux)*nedg);

    for(i=0; i<nedg; i++)
    {      
      if (edges[i].cordomtest(lnn))
      {
        aux[i]++;
        nedg_l++;
      }
    }
    fprintf(out, "edges %ld\n", nedg_l);
    for(i=0; i<nedg; i++)
    {
      if (aux[i])
        edges[i].print(out, lnn);
    }
    delete [] aux;
  }
}



/**
  The function searches the list of property numbers for the given number aprop.
  It returns whether the aprop is a new property number.

  @param aprop - searched property number
  @param propid - array of with list of property numbers
  @param n      - length of array propid
  
  @retval 0 - the property was found in the list
  @retval 1 - the property was not found in the list

  Created by Tomas Koudelka, 08.2011
*/
long sedges::search_list_newprop(long aprop, long *propid, long n)
{
  long i;

  for(i=0; i<n; i++)
  {
    if (propid[i] == aprop)
      return 0;
  }
  return 1;
}



/**
  The function generates from the list of additional edges new lists which contain 
  the list of different property numbers, lists of nodes for particular property numbers, 
  lists of elements for particular property numbers and lists of edge id for particular
  lists of elements with given property number.

  @param nn - total number of nodes in the mesh
  @param ne - total number of elements in the mesh
  @param elems - array of all elements in the mesh

  @return The function does not return anything.

  Created by Tomas Koudelka, 08.2011
*/
void sedges::gen_list_edg(long nn, long ne, selement *elems)
{
  long i;
  long aux;
  long *propid;


  if (nedg > 0)
  {
    propid = new long[nedg];
    nprop = 0;
    propid[nprop] = aux = edges[0].prop;
    nprop++;
    for(i=0; i<nedg; i++)
    {
      if (edges[i].prop != aux)
      { 
        aux = edges[i].prop; 
        if (search_list_newprop(aux, propid, nprop))
        {
          propid[nprop] = aux;
          nprop++;
        }
      }
    }    
    proplist = new long[nprop];
    for(i=0; i<nprop; i++)
      proplist[i] = propid[i];
    delete [] propid;
    prop_list_nod       = new long*[nprop];
    prop_list_nod_num   = new long[nprop];
    prop_list_elem      = new long*[nprop];
    prop_list_elem_num  = new long[nprop];
    prop_list_elem_edgid = new long**[nprop];
    prop_list_elem_edgid_num = new long*[nprop];
    memset(prop_list_nod, 0, sizeof(*prop_list_nod)*nprop);
    memset(prop_list_nod_num, 0, sizeof(*prop_list_nod_num)*nprop);
    memset(prop_list_elem, 0, sizeof(*prop_list_elem)*nprop);
    memset(prop_list_elem_num, 0, sizeof(*prop_list_elem_num)*nprop);
    memset(prop_list_elem_edgid, 0, sizeof(*prop_list_elem_edgid)*nprop);
    memset(prop_list_elem_edgid_num, 0, sizeof(*prop_list_elem_edgid_num)*nprop);

    gen_list_node(nn);
    gen_list_elem(ne, elems);
  }
}



/**
  The function generates from the list of additional edges new list which contain 
  lists of nodes for particular different property numbers. The lists are stored
  in the array prop_list_nod where prop_list_nod[i] represents array of
  node numbers for the property number proplist[i]. The length of array
  prop_list_nod[i] is given by prop_list_nod_num[i].

  @param nn - total number of nodes in the mesh

  @return The function does not return anything.

  Created by Tomas Koudelka, 08.2011
*/
void sedges::gen_list_node(long nn)
{
  long i, j, k, nnodp;
  long *nodid;

  nodid = new long[nn];

  for(i=0; i<nprop; i++)
  {
    memset(nodid, 0, sizeof(*nodid)*nn);   
    nnodp = 0;
    for(j=0; j<nedg; j++)
    {
      if (edges[j].prop == proplist[i])
      {
        if (nodid[edges[j].n1] == 0)
        {
          nodid[edges[j].n1] = 1;
          nnodp++;
        }
        if (nodid[edges[j].n2] == 0)
        {
          nodid[edges[j].n2] = 1;
          nnodp++;
        }
      }
    }
    prop_list_nod_num[i] = nnodp;
    prop_list_nod[i] = new long[nnodp];
    k = 0;
    for(j=0; j<nn; j++)
    {
      if (nodid[j])
      {
        prop_list_nod[i][k] = j;
        k++;
      }
    }
  }
  delete [] nodid;
}



/**
  The function generates from the list of additional edges new lists which contain 
  lists of elements for particular different property numbers. The lists are stored
  in the array prop_list_nod where prop_list_elem[i] represents array of
  element numbers for the property number proplist[i]. The length of array
  prop_list_elem[i] is given by prop_list_elem_num[i]. 

  Additionally, the function  creates lists of edge indeces with given property number 
  for each list of elements. Edge indeces are stored in the 
  array prop_list_elem_edgid where the prop_list_elem_edgid[i][j] represents array of 
  edge indeces with property number proplist[i] for the j-th element from the prop_list_elem[i].
  The length of array  prop_list_elem_edgid[i] is given by prop_list_elem_edgid_num[i][j].    

  @param ne - total number of elements in the mesh
  @param elems - array of all elements in the mesh

  @return The function does not return anything.

  Created by Tomas Koudelka, 08.2011
*/
void sedges::gen_list_elem(long ne, selement *elems)
{
  long i, j, k, l, nelemp;
  long *elemid;
  long **elemedgid;
  long edgid;
  long nodes[2];
  long aux;

  elemid    = new long[ne];
  elemedgid = new long*[ne];
  for(i=0; i<ne; i++)
    elemedgid[i] = new long [elems[i].ned];

  for(i=0; i<nprop; i++)
  {
    memset(elemid, 0, sizeof(*elemid)*ne);
    for(j=0; j<ne; j++)
      memset(elemedgid[j], 0, sizeof(*elemedgid[j])*elems[j].ned);
    nelemp = 0;

    for(j=0; j<nedg; j++)
    {
      if (edges[j].prop == proplist[i])
      {        
        for(k=0; k<ne; k++)
        {
          nodes[0] = edges[j].n1;
          nodes[1] = edges[j].n2;
          edgid = elems[k].compare_edg(nodes, 2);
          if (edgid >= 0)
          {
            if (elemid[k] == 0)
            {
              nelemp++;
              elemid[k] = 1;
            }
            elemedgid[k][edgid] = 1;
          }
        }
      }
    }
    prop_list_elem_num[i]      = nelemp;
    prop_list_elem[i]          = new long[nelemp];
    prop_list_elem_edgid_num[i] = new long[nelemp];
    prop_list_elem_edgid[i]     = new long*[nelemp];
    k = 0;
    for(j=0; j<ne; j++)
    {
      if (elemid[j] == 0)
        continue;

      prop_list_elem[i][k] = j;
      aux = 0;
      for(l = 0; l<elems[j].ned; l++)
      {
        if (elemedgid[j][l])
          aux++;
      }
      prop_list_elem_edgid_num[i][k] = aux;
      if (aux)
        prop_list_elem_edgid[i][k] = new long[aux];
      else
        prop_list_elem_edgid[i][k] = NULL;
      aux = 0;
      for(l = 0; l<elems[j].ned; l++)
      {
        if (elemedgid[j][l])
        {
          prop_list_elem_edgid[i][k][aux] = l;
          aux++;
        }
      }
      k++;
    }
  }
  delete [] elemid;
  for(i=0; i<ne; i++)
    delete [] elemedgid[i];
  delete [] elemedgid;
}



/**
  The constructor initializes all data members to -1, 0 or NULL values.

  Created by Tomas Koudelka, 08.2011
*/
sface::sface(void)
{
  nnod = 0;
  nodes = NULL;
  prop = -1;
}



/**
  The constructor initializes all data members to -1 or 0 values.
  It allocates memory for n nodes.

  @param - the number of nodes defining one face

  Created by Tomas Koudelka, 08.2011
*/
sface::sface(long n)
{
  nnod = n;
  nodes = new long[n];
  memset(nodes, 0, sizeof(*nodes)*n);
  prop = -1;
}



/**
  The destructor releases allocated memory.

  Created by Tomas Koudelka, 08.2011
*/
sface::~sface(void)
{
  delete [] nodes;
}



/**
  The function reads data about one face from the opened text file.

  @param in  - pointer to the opened text file

  @retval 0 - on success
  @retval 1 - in the case of an read error

  Created by Tomas Koudelka, 08.2011
*/
long sface::read(XFILE *in)
{
  long i;

  if (xfscanf(in, "%ld", &nnod) == 0)
    return 1;

  nodes = new long[nnod];
  memset(nodes, 0, sizeof(*nodes)*nnod);

  for(i=0; i<nnod; i++)
  {
    if (xfscanf(in, "%ld", nodes+i) == 0)
      return 1;
    // change to C/C++ index notation
    nodes[i]--;

  }
  if (xfscanf(in, "%ld", &prop) == 0)
    return 1;

  return 0;
}



/**
  The function prints data about one face to the opened text file.
  If the second parameter lnn is not NULL, 
  then the node numbers on the face are renumbered using this array.
  The lnn array is used in connection with the mesh decompoistion and it
  contains local node numbers for a domain. The size of the array is given
  by the total number of nodes in the problem.

  @param in - pointer to the opened text file
  @param lnn - pointer to array of local node numbers used for the
               mesh decomposition. Default value of this parameter is NULL.

  @return The function does not return anything.

  Created by Tomas Koudelka, 08.2011
*/
void sface::print(FILE *out, long *lnn)
{
  long i;
  
  fprintf(out, "%ld", nnod);
  for(i=0; i<nnod; i++)
  {
    if (lnn)
      fprintf(out, " %ld", lnn[nodes[i]]);
    else
      fprintf(out, " %ld", nodes[i]+1);
  }

  fprintf(out, " %ld\n", prop);
}



/**
  The function tests whether all nodes of the face are contained
  in the given subdomain whose local node numbers given in 
  the parameter lnn. The lnn array is used in connection with
  the mesh decompoistion and it contains local node numbers for a domain. The size 
  of the array is given by the total number of nodes in the problem.

  @param lnn - pointer to array of local node numbers used for the
               mesh decomposition. Default value of this parameter is NULL.

  @retval 0 - face is not contained in the given subdomain
  @retval 1 - face is contained in the given subdomain

  Created by Tomas Koudelka, 09.2011
*/
long sface::cordomtest(long *lnn)
{
  long i, aux=0;

  for(i=0; i<nnod; i++)
  {
    if (lnn[nodes[i]])
     aux++;
  }
  if (aux == nnod)
    return 1;

  return 0;
}



/**
  The constructor initializes all data members to zero or NULL values.

  Created by Tomas Koudelka, 08.2011
*/
sfaces::sfaces()
{
  nfcs = 0L;
  nprop = 0L;
  faces = NULL;
  proplist = NULL;
  prop_list_nod = NULL;
  prop_list_nod_num = NULL;
  prop_list_elem = NULL;
  prop_list_elem_num = NULL;
  prop_list_elem_sfid = NULL;
  prop_list_elem_sfid_num = NULL;
}



/**
  The destructor releases all allocated memory.

  Created by Tomas Koudelka, 08.2011
*/
sfaces::~sfaces()
{
  long i, j;

  delete [] faces;
  delete [] proplist;
  for (i=0; i<nprop; i++)
  {
    for (j=0; j<prop_list_elem_num[i]; j++)
      delete [] prop_list_elem_sfid[i][j];
    
    delete [] prop_list_nod[i];
    delete [] prop_list_elem[i];
    delete [] prop_list_elem_sfid[i];
    delete [] prop_list_elem_sfid_num[i];
  }
  if (nprop)
  {
    delete [] prop_list_nod;
    delete [] prop_list_nod_num;
    delete [] prop_list_elem;
    delete [] prop_list_elem_num;
    delete [] prop_list_elem_sfid;
    delete [] prop_list_elem_sfid_num;
  }
}



/**
  The constructor allocates memory for the list of faces
  and initializes rest of data members to zero or NULL values.
  
  @param n - number of allocated edges

  Created by Tomas Koudelka, 08.2011
*/
sfaces::sfaces(long n)
{
  nfcs = n;
  nprop = 0L;
  faces = new sface[n];
  proplist = NULL;
  prop_list_nod = NULL;
  prop_list_nod_num = NULL;
  prop_list_elem = NULL;
  prop_list_elem_num = NULL;
  prop_list_elem_sfid = NULL;
  prop_list_elem_sfid_num = NULL;
}



/**
  The function reads list of additional surfaces from the opened text file.
  The list is placed at the end of file with topology and it is created by
  MeshEditor usually.

  @param in - pointer to the opened text file

  @return The function does not return anything.

  Created by Tomas Koudelka, 08.2011
*/
void sfaces::read(XFILE *in)
{
  long i;

  for(i=0; i<nfcs; i++)
  {
    if (faces[i].read(in))
    {
      print_err("cannot read face number %ld\n", __FILE__, __LINE__, __func__, i+1);
      abort();
    }
  }
}



/**
  The function prints list of additional surfaces from the opened text file.
  The list is placed the end of file with topology usually in MeshEditor format.
  If the second parameter lnn is not NULL, then only the faces whose all local node numbers
  are nonzero are printed. The lnn array is used in connection with
  the mesh decompoistion and it contains local node numbers for a domain. The size 
  of the array is given by the total number of nodes in the problem.
  

  @param out - pointer to the opened text file
  @param lnn - pointer to array of local node numbers used for the
               mesh decomposition. Default value of this parameter is NULL.

  @return The function does not return anything.

  Created by Tomas Koudelka, 08.2011
*/
void sfaces::print(FILE *out, long *lnn)
{
  long i;
  long *aux;
  long nfcs_l;
  
  if (lnn == NULL) // no mesh decomposition
  {
    fprintf(out, "faces %ld\n", nfcs);
    for(i=0; i<nfcs; i++)
      faces[i].print(out, lnn);
  }
  else
  {
    nfcs_l=0;
    aux = new long[nfcs];
    memset(aux, 0, sizeof(*aux)*nfcs);

    for(i=0; i<nfcs; i++)
    {      
      if (faces[i].cordomtest(lnn))
      {
        aux[i]++;
        nfcs_l++;
      }
    }
    fprintf(out, "faces %ld\n", nfcs_l);
    for(i=0; i<nfcs; i++)
    {
      if (aux[i])
        faces[i].print(out, lnn);
    }
    delete [] aux;
  }
}



/**
  The function searches the list of property numbers for the given number aprop.
  It returns whether the aprop is a new property number.

  @param aprop - searched property number
  @param propid - array of with list of property numbers
  @param n      - length of array propid
  
  @retval 0 - the property was found in the list
  @retval 1 - the property was not found in the list

  Created by Tomas Koudelka, 08.2011
*/
long sfaces::search_list_newprop(long prop, long *propid, long n)
{
  long i;

  for(i=0; i<n; i++)
  {
    if (propid[i] == prop)
      return 0;
  }
  return 1;
}



/**
  The function generates from the list of additional surfaces new lists which contain 
  the list of different property numbers, lists of nodes for particular property numbers, 
  lists of elements for particular property numbers and lists of surface id for particular
  lists of elements with given property number.

  @param nn - total number of nodes in the mesh
  @param ne - total number of elements in the mesh
  @param elems - array of all elements in the mesh

  @return The function does not return anything.

  Created by Tomas Koudelka, 08.2011
*/
void sfaces::gen_list_surf(long nn, long ne, selement *elems)
{
  long i;
  long aux;
  long *propid;


  if (nfcs > 0)
  {
    propid = new long[nfcs];
    nprop = 0;
    propid[nprop] = aux = faces[0].prop;
    nprop++;
    for(i=0; i<nfcs; i++)
    {
      if (faces[i].prop != aux)
      { 
        aux = faces[i].prop; 
        if (search_list_newprop(aux, propid, nprop))
        {
          propid[nprop] = aux;
          nprop++;
        }
      }
    }    
    proplist = new long[nprop];
    for(i=0; i<nprop; i++)
      proplist[i] = propid[i];
    delete [] propid;
    prop_list_nod           = new long*[nprop];
    prop_list_nod_num       = new long[nprop];
    prop_list_elem          = new long*[nprop];
    prop_list_elem_num      = new long[nprop];
    prop_list_elem_sfid     = new long**[nprop];
    prop_list_elem_sfid_num = new long*[nprop];
    memset(prop_list_nod, 0, sizeof(*prop_list_nod)*nprop);
    memset(prop_list_nod_num, 0, sizeof(*prop_list_nod_num)*nprop);
    memset(prop_list_elem, 0, sizeof(*prop_list_elem)*nprop);
    memset(prop_list_elem_num, 0, sizeof(*prop_list_elem_num)*nprop);
    memset(prop_list_elem_sfid, 0, sizeof(*prop_list_elem_sfid)*nprop);
    memset(prop_list_elem_sfid_num, 0, sizeof(*prop_list_elem_sfid_num)*nprop);

    gen_list_node(nn);
    gen_list_elem(ne, elems);
  }
}



/**
  The function generates from the list of additional surfaces new list which contain 
  lists of nodes for particular different property numbers. The lists are stored
  in the array prop_list_nod where prop_list_nod[i] represents array of
  node numbers for the property number proplist[i]. The length of array
  prop_list_nod[i] is given by prop_list_nod_num[i].

  @param nn - total number of nodes in the mesh

  @return The function does not return anything.

  Created by Tomas Koudelka, 08.2011
*/
void sfaces::gen_list_node(long nn)
{
  long i, j, k, nnodp;
  long *nodid;

  nodid = new long[nn];

  for(i=0; i<nprop; i++)
  {
    memset(nodid, 0, sizeof(*nodid)*nn);   
    nnodp = 0;
    for(j=0; j<nfcs; j++)
    {
      if (faces[j].prop == proplist[i])
      {
        for(k=0; k<faces[j].nnod; k++)
        {
          if (nodid[faces[j].nodes[k]] == 0)
          {
            nodid[faces[j].nodes[k]] = 1;
            nnodp++;
          }
        }
      }
    }
    prop_list_nod_num[i] = nnodp;
    prop_list_nod[i] = new long[nnodp];
    k = 0;
    for(j=0; j<nn; j++)
    {
      if (nodid[j])
      {
        prop_list_nod[i][k] = j;
        k++;
      }
    }
  }
  delete [] nodid;
}



/**
  The function generates from the list of additional surfaces new lists which contain 
  lists of elements for particular different property numbers. The lists are stored
  in the array prop_list_nod where prop_list_elem[i] represents array of
  element numbers for the property number proplist[i]. The length of array
  prop_list_elem[i] is given by prop_list_elem_num[i]. 

  Additionally, the function  creates lists of surface indeces with given property number 
  for each list of elements. Surface indeces are stored in the 
  array prop_list_elem_sfid where the prop_list_elem_sfid[i][j] represents array of 
  surface indeces with property number proplist[i] for the j-th element from the prop_list_elem[i].
  The length of array  prop_list_elem_sfid[i] is given by prop_list_elem_sfid_num[i][j].    

  @param ne - total number of elements in the mesh
  @param elems - array of all elements in the mesh

  @return The function does not return anything.

  Created by Tomas Koudelka, 08.2011
*/
void sfaces::gen_list_elem(long ne, selement *elems)
{
  long i, j, k, l, nelemp;
  long *elemid;
  long **elemsfcid;
  long sfid, aux;

  elemid    = new long[ne];
  elemsfcid = new long*[ne];
  for(i=0; i<ne; i++)
    elemsfcid[i] = new long [elems[i].nsurf];

  for(i=0; i<nprop; i++)
  {
    memset(elemid, 0, sizeof(*elemid)*ne);
    for(j=0; j<ne; j++)
      memset(elemsfcid[j], 0, sizeof(*elemsfcid[j])*elems[j].nsurf);
    nelemp = 0;

    for(j=0; j<nfcs; j++)
    {
      if (faces[j].prop == proplist[i])
      {        
        for(k=0; k<ne; k++)
        {
          sfid = elems[k].compare_surf(faces[j].nodes, faces[j].nnod);
          if (sfid >= 0)
          {
            if (elemid[k] == 0)
            {
              nelemp++;
              elemid[k] = 1;
            }
            elemsfcid[k][sfid] = 1;
          }
        }
      }
    }
    prop_list_elem_num[i]      = nelemp;
    prop_list_elem[i]          = new long[nelemp];
    prop_list_elem_sfid_num[i] = new long[nelemp];
    prop_list_elem_sfid[i]     = new long*[nelemp];
    k = 0;
    for(j=0; j<ne; j++)
    {
      if (elemid[j] == 0)
        continue;

      prop_list_elem[i][k] = j;
      aux = 0;
      for(l = 0; l<elems[j].nsurf; l++)
      {
        if (elemsfcid[j][l])
          aux++;
      }
      prop_list_elem_sfid_num[i][k] = aux;
      if (aux)
        prop_list_elem_sfid[i][k] = new long[aux];
      else
        prop_list_elem_sfid[i][k] = NULL;
      aux = 0;
      for(l = 0; l<elems[j].nsurf; l++)
      {
        if (elemsfcid[j][l])
        {
          prop_list_elem_sfid[i][k][aux] = l;
          aux++;
        }
      }
      k++;
    }
  }
  delete [] elemid;
  for(i=0; i<ne; i++)
    delete [] elemsfcid[i];
  delete [] elemsfcid;
}



/**
  This constructor initializes data mebers with NULL or zero values

  Created by Tomas Koudelka 03/2008, koudelka@cml.fsv.cvut.cz
*/
siftop :: siftop (void)
{
  elements = NULL;
  nodes    = NULL;
  gnn      = NULL;
  edges    = NULL;
  surfaces = NULL;
  nn       = 0;
  ne       = 0;
  memset(npet, 0, sizeof(npet));
}



/**
  This destructor deallocates memory used by data mebers

  Created by Tomas Koudelka 03.2008, koudelka@cml.fsv.cvut.cz
*/
siftop :: ~siftop(void)
{
  if (elements)
    delete [] elements;

  if (nodes)
    delete [] nodes;

  if (gnn)
    delete [] gnn;

  if (edges)
    delete edges;

  if (surfaces)
    delete surfaces;
}



/**
  This method reads topology of given domain from file, which is in the JKTK format

  @param in   - pointer to structure with opened file of input data
  @param rgnn - indicator whether global node numbers array should be read (=1) or no (=0).
                This is support for paralell computing.

  @retval 0 : on success
  @retval 1 : on error in reading of node
  @retval 2 : on error in reading of element
  @retval 3 : on error in reading of global node numbers

  @note JKTK mesh format is generated by programs gensif??? and details are in . . .

  created by Tomas Koudelka 3.2008, koudelka@cml.fsv.cvut.cz
*/
long siftop::read(XFILE *in, long rgnn, long rebp)
{
  long i, j, eid, nid;
  long nfaces, nedges;
  char emsg[1001];
  
  //  number of nodes
  xfscanf(in, "%k%ld", "num_nodes", &nn);
  nodes = new snode[nn];
  for (i = 0; i < nn; i++)
  {
    xfscanf(in, "%k%ld", "node_id", &nid);
    if ((nid < 1) && (nid > nn))
    {
      sprintf(emsg, "node number %ld is out of range <1, %ld> (file %s, line=%ld, col=%ld)\n", nid, nn, in->fname, in->line, in->col);
      print_err(emsg, __FILE__, __LINE__, __func__);
      return (1);
    }
    nid--;
    xfscanf(in, "%k%le %k%le %k%le %k%ld", "x", &nodes[nid].x, "y", &nodes[nid].y, "z", &nodes[nid].z, "numprop", &nodes[nid].nprop);
    if (nodes[nid].nprop < 1)
    {
      sprintf(emsg, "node %ld must have at least one property (file %s, line=%ld, col=%ld)\n", nid, in->fname, in->line, in->col);
      print_err(emsg, __FILE__, __LINE__, __func__);
      return (1);
    }
    nodes[nid].alloc(nodes[nid].nprop);
    for (j=0; j < nodes[nid].nprop; j++)
      xfscanf(in, "%k%m%ld", "prop", &entityp_kwdset, nodes[nid].entid+j, nodes[nid].prop+j);
  }


  //  number of elements
  xfscanf(in, "%k%ld", "num_elements", &ne);
  elements = new selement[ne];
  for (i = 0; i < ne; i++)
  {
    xfscanf(in, "%k%ld", "elem_id", &eid);
    eid--;
    if ((eid < 0) && (eid >= ne))
    {
      sprintf(emsg, "element number %ld is out of range <1, %ld> (file %s, line=%ld, col=%ld)", eid+1, nn, in->fname, in->line, in->col);
      print_err(emsg, __FILE__, __LINE__, __func__);
      return (2);
    }
    xfscanf(in, "%k%m", "eltype", &gtypel_kwdset, &elements[eid].type);
    if (elements[eid].alloc(rebp))
    {
      sprintf(emsg, "unknown type on element %ld is required (file %s, line=%ld, col=%ld)", eid+1, in->fname, in->line, in->col);
      print_err(emsg, __FILE__, __LINE__, __func__);
    }
    xfscanf(in, "%k", "enodes");
    for (j = 0; j < elements[eid].nne; j++)
    {
      xfscanf(in, "%ld", &elements[eid].nodes[j]);
      elements[eid].nodes[j]--;
    }
    xfscanf(in, "%k%ld", "eprop", &elements[eid].prop);
    if (rebp)
    {
      if (elements[eid].ned)
        xfscanf(in, "%k", "propedg");
      for(j = 0; j < elements[eid].ned; j++)
        xfscanf(in, "%ld", elements[eid].propedg+j);
      if (elements[eid].nsurf)
        xfscanf(in, "%k", "propsurf");
      for(j = 0; j < elements[eid].nsurf; j++)
        xfscanf(in, "%ld", elements[eid].propsurf+j);
    }
    else{ // skip rest of line
      xfscanf(in, "%*[^\n]");
    }
  }

  
  if (rgnn==1){
    //  parallel code
    gnn = new long[nn];
    memset(gnn, 0, sizeof(*gnn)*nn);
    for (i = 0; i < nn; i++){
      xfscanf(in, "%k%ld", "node_id", &j);
      if ((j < 1) && (j > nn)){
	sprintf(emsg, "node number %ld is out of range <1, %ld> (file %s, line=%ld, col=%ld)\n", j, nn, in->fname, in->line, in->col);
	print_err(emsg, __FILE__, __LINE__, __func__);
	return (3);
      }
      j--;
      xfscanf(in, "%k%ld", "glob_id", gnn+j);
    }
  }

  if (rgnn==2){
    //  sequential code, BOSS method or sequential version of the FETI method

    //  begin, added, JK, 28.10.2009
    
    //  mesh description
    //  number of subdomains/aggregates
    xfscanf (in,"%ld %ld",&meshtype,&nsd);
    
    //  number of nodes on subdomains/aggregates
    nnsd = new long [nsd];
    //  loop over the number of subdomains/aggregates
    for (i=0;i<nsd;i++){
      xfscanf (in,"%ld",nnsd+i);
    }
    //  end, JK, 28.10.2009
    
    gnn = new long[nn];
    memset(gnn, 0, sizeof(*gnn)*nn);
    for (i = 0; i < nn; i++){
      xfscanf(in, "%k%ld", "node_id", &j);
      if ((j < 1) && (j > nn)){
	sprintf(emsg, "node number %ld is out of range <1, %ld> (file %s, line=%ld, col=%ld)\n", j, nn, in->fname, in->line, in->col);
	print_err(emsg, __FILE__, __LINE__, __func__);
	return (3);
      }
      j--;
      xfscanf(in, "%k%ld", "glob_id", gnn+j);
    }
  }

  kwd_handling bmode = in->kwdmode;
  in->kwdmode = sequent_mode;
  // read additional data about surfaces created in the MeshEditor
  if (xfscanf(in, " %+k", "faces"))
  {
    in->kwdmode = bmode;
    xfscanf(in, "%ld", &nfaces);  
    surfaces = new sfaces(nfaces);
    surfaces->read(in); 
    surfaces->gen_list_surf(nn, ne, elements);
  }
  
  in->kwdmode = sequent_mode;
  // read additional data about edges created in the MeshEditor
  if (xfscanf(in, " %+k", "edges"))
  {
    in->kwdmode = bmode;
    xfscanf(in, "%ld", &nedges);
    edges = new sedges(nedges);
    edges->read(in); 
    edges->gen_list_edg(nn, ne, elements);
  }
  in->kwdmode = bmode;
    
  return(0);
}



/**
  This method copies mesh topology to general topology structure.
  It is used in preprocessors especially.

  @param top  - pointer to structure of general topology

  created by Tomas Koudelka 3.2008, koudelka@cml.fsv.cvut.cz
*/
void siftop::exporttop (gtopology *top)
{
  long i, j;
  top->alloc_nodes(nn);
  top->alloc_elements(ne);
  for (i = 0; i < nn; i++)
  {
    top->gnodes[i].x = nodes[i].x;
    top->gnodes[i].y = nodes[i].y;
    top->gnodes[i].z = nodes[i].z;
  }
  for (i = 0; i < ne; i++)
  {
    top->gelements[i].nne = elements[i].nne;
    top->gelements[i].nodes = new long[elements[i].nne];
    for (j = 0; j < elements[i].nne; j++)
      top->gelements[i].nodes[j] = elements[i].nodes[j];
    switch(elements[i].type){
    case isolinear1d:{
      top->gelements[i].get=linbar;
      break;
    }
    case isoquadratic1d:{
      top->gelements[i].get=quadbar;
      break;
    }
    case trianglelinear:{
      top->gelements[i].get=lintriag;
      break;
    }
    case trianglequadratic:{
      top->gelements[i].get=quadtriag;
      break;
    }
    case isolinear2d:{
      top->gelements[i].get=linquad;
      break;
    }
    case isoquadratic2d:{
      top->gelements[i].get=quadquad;
      break;
    } 
    case tetrahedronlinear:{
      top->gelements[i].get=lintetra;
      break;
    }
    case tetrahedronquadratic:{
      top->gelements[i].get=quadtetra;
      break;
    }
    case pyramidelinear:{
      top->gelements[i].get=noelem;
      break;
    }
    case pyramidequadratic:{
      top->gelements[i].get=noelem;
      break;
    }
    case vedgelinear:{
      top->gelements[i].get=noelem;
      break;
    }
    case vedgequadratic:{
      top->gelements[i].get=noelem;
      break;
    }
    case isolinear3d:{
      top->gelements[i].get=linhexa;
      break;
    }
    case isoquadratic3d:{
      top->gelements[i].get=quadhexa;
      break;
    }
    case all_element_types:{
      top->gelements[i].get=noelem;
      break;
    }
    }
  }
  
}



/**
  This function imports tpology in T3D format from the opened file in.

  @param in - pointer to structure with opened file of input data
  @param paral - flag for topology format used in parallel computations

  @retval -1 if topology pointer S is not set
  @retval -2 if file specified by ofname cannot be opened

  created  03.2008, Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
void siftop::import_t3d (XFILE *in, long paral)
{
  long i;
  long type, renum, mode, temp;
  long outmode[9];


  xfscanf(in, "%ld", &type);   // reads type of generator
  xfscanf(in, "%ld", &degree); // reads degrees of element aproximation
  xfscanf(in, "%ld", &renum);  // reads indicator of renumbering
  xfscanf(in, "%ld", &mode);   // reads output mode for elements and nodes
    
  if (paral==1)
    xfscanf(in,"%ld", &temp);
    
  // decodes mode variable and stores results in outmode array
  for (i = 0; i < 9; i++)
    outmode[i]=(mode>>i) & 1;

  xfscanf(in, "%ld", &nn);
  // allocation of memory for nodes
  if (nodes)
    delete [] nodes; 
  nodes = new snode[nn];
  if (paral)  // mesh for parallel computations is required 
  {
    if (gnn)
      delete [] gnn; 
    gnn = new long[nn];
    memset(gnn, 0, sizeof(*gnn)*nn);
  }
  switch (type) 
  // number of particular element types depends on mesh generator type 
  {
    case 3:
      // reading of number of edges, trias, tetras
      xfscanf(in, "%ld%ld%ld", npet+isolinear1d+degree-2, npet+trianglelinear+degree-2, npet+tetrahedronlinear+degree-2);
      break ;
    case 4:
      // reading of number of edges, quads and bricks
      xfscanf(in, "%ld%ld%ld", npet+isolinear1d+degree-2, npet+isolinear2d+degree-2, npet+isolinear3d+degree-2);
      break ;
    case 7:
      // reading of number of edges, trias, quads
      xfscanf(in, "%ld%ld%ld",  npet+isolinear1d+degree-2, npet+trianglelinear+degree-2, npet+isolinear2d+degree-2);
      // reading of number of tetras, pyrams, vedges and bricks
      xfscanf(in, "%ld%ld%ld%ld", npet+tetrahedronlinear+degree-2, npet+pyramidelinear+degree-2, npet+vedgelinear+degree-2, npet+isolinear3d+degree-2);
      break ;
    default:
      print_err("unknown type of generator is required", __FILE__, __LINE__, __func__);
  }
  // total number of elements
  ne=0;
  for (i = 0; i < all_element_types-1; i++)
    ne+=npet[i];

  // import of section of nodes
  import_t3d_nodes(in, paral, outmode) ;

  // allocation of memory for elements
  if (elements)
    delete [] elements;
  elements = new selement[ne];

  // import of particular element types 
  import_t3d_elements(in, outmode);

  return;
}



/**
  Function imports section of nodes in the T3D format.

  @param in - pointer to structure with opened file of input data
  @param paral   - indicates whether global node numbers are read (=1) or not (=0).
  @param outmode - array with description of output type (see T3D documentation).
                   Each aray element represents one bit flag from T3D output type.

  @retval 0 : on succes
  @retval 1 : in case that node number is out of range

  created  03.2008, Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
long siftop::import_t3d_nodes (XFILE *in, long paral, long *outmode)
{
  long i, j, nid;
  long nent;
  entityp ent;
  long eprop;
  char emsg[1001];

  for (i = 0; i < nn; i++)
  {
    xfscanf(in, "%ld", &nid);
    if ((nid < 1) || (nid > nn))
    {
      sprintf(emsg, "node number %ld is out of range <1, %ld> (file %s, line=%ld, col=%ld)\n", nid, nn, in->fname, in->line, in->col);
      print_err(emsg, __FILE__, __LINE__, __func__);
      return (1);
    }
    nid--;
    if (paral)
      xfscanf(in, "%ld", gnn+nid);
    xfscanf(in, "%le %le %le %m %*ld %ld", &nodes[nid].x, &nodes[nid].y, &nodes[nid].z, &entityp_kwdset, &ent, &eprop);
    if (outmode[omt_sec_clasif_nod])
    {
      xfscanf(in, "%ld", &nent);
      nent++;
      nodes[nid].alloc(nent);
      nodes[nid].entid[0] = ent;
      nodes[nid].prop[0]  = eprop;
      for (j=1; j < nent; j++)
        xfscanf(in, "%m %*ld %ld", &entityp_kwdset, nodes[nid].entid+j, nodes[nid].prop+j);
    }
    else
    {
      nent = 1;
      nodes[nid].alloc(nent);
      nodes[nid].entid[0] = ent;
      nodes[nid].prop[0]  = eprop;
    }
    switch (nodes[nid].entid[0])
    {
      case evertex:
      case eregion:
        break ;
      case ecurve:
        if (outmode[omt_parameters] || outmode[omt_tangent])
          skipline(in); // skip rest of actual line
        break ;
      case esurface:
      case epatch:
      case eshell:
        if (outmode[omt_parameters] || outmode[omt_normal])
          skipline(in); // skip rest of actual line
        break ;
    }
  }
  return(0);
}



/**
  This function imports all elements from T3D file.

  @param in - pointer to structure with opened file of input data
  @param outmode - array with description of output type (see T3D documentation).
                   Each aray element represents one bit flag from T3D output type.
  @retval 0 : on success
  @retval 1 : in case that element number is out of range

  Created 03.2008, Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
long siftop::import_t3d_elements (XFILE *in, long *outmode)
{
  long i, j, eid;
  gtypel etype;
  char emsg[1001];
  
  if (degree == 1) 
    etype=isolinear1d;
  else
    etype=isoquadratic1d;

  for ( ; etype<all_element_types; etype++, etype++)
  {
    for (i=0; i<npet[etype-1]; i++)
    {
      xfscanf(in, "%ld", &eid);
      eid--;
      if ((eid < 0) || (eid >= ne))
      {
        sprintf(emsg, "element number %ld is out of range <1, %ld> (file %s, line=%ld, col=%ld)", eid+1, nn, in->fname, in->line, in->col);
        print_err(emsg, __FILE__, __LINE__, __func__);
        return(1);
      }

      elements[eid].type = etype;
      if (outmode[omt_neighbourhood])
        elements[eid].alloc(1);
      else
        elements[eid].alloc(1);

      for(j=0; j<elements[eid].nne; j++)
      {
        xfscanf(in, "%ld", &elements[eid].nodes[j]);
        elements[eid].nodes[j]--;
      }

      xfscanf(in, "%*ld %*ld %ld", &elements[eid].prop);
      switch (etype)
      {
        case isolinear1d:
        case isoquadratic1d:
          break;
        case trianglelinear:
        case trianglequadratic:
        case isolinear2d:
        case isoquadratic2d:
        {  
          if (outmode[omt_neighbourhood])
            skipline(in); // skip rest of actual line
          if (outmode[omt_boundary_entities])
          {
            for (j=0; j<elements[eid].ned; j++)
              xfscanf(in, "%*ld");
            for (j=0; j<elements[eid].ned; j++)
              xfscanf(in, "%ld", elements[eid].propedg+j);
            skipline(in); // skip rest of actual line
          }
          else
            skipline(in); // skip rest of actual line
          break;
        }
        case tetrahedronlinear:
        case tetrahedronquadratic:
        case pyramidelinear:
        case pyramidequadratic:
        case vedgelinear:
        case vedgequadratic:
        case isolinear3d:
        case isoquadratic3d:
        {  
          if (outmode[omt_neighbourhood] || outmode[omt_associated_elements])
            skipline(in); // skip rest of actual line
          if (outmode[omt_boundary_entities])
          {
            for (j=0; j<elements[eid].nsurf; j++)
              xfscanf(in, "%*ld"); // skip entity id
            for (j=0; j<elements[eid].nsurf; j++)
              xfscanf(in, "%*ld"); // skip entity types
            for (j=0; j<elements[eid].nsurf; j++)
              xfscanf(in, "%ld", elements[eid].propsurf+j);
            skipline(in); // skip rest of actual line
          }
          else
            skipline(in); // skip rest of actual line
          break;
        }
        default:
          sprintf(emsg, "unknown type on element %ld is required (file %s, line=%ld, col=%ld)", eid+1, in->fname, in->line, in->col);
          print_err(emsg, __FILE__, __LINE__, __func__);
          break;
      }
    }
  }
  return(0);
}



/**
   Function prints mesh data in SSMF.
   
   created by Tomas Koudelka 3.2008, koudelka@cml.fsv.cvut.cz
*/
void siftop::print (FILE *out)
{
  long i, j;
  
  fprintf (out,"%ld\n",nn);

  for (i = 0; i < nn; i++)
  {
    fprintf (out,"%ld %15.10le %15.10le %15.10le %ld",i+1, nodes[i].x, nodes[i].y, nodes[i].z, nodes[i].nprop);
    for(j=0; j<nodes[i].nprop; j++)
    {
      fprintf(out, " %d %ld", nodes[i].entid[j], nodes[i].prop[j]);
    }
    fprintf(out, "\n");
  }
  
  fprintf (out,"\n%ld\n", ne);
  
  for (i = 0; i < ne; i++)
  {
    fprintf (out,"%ld %d",i+1, elements[i].type);
    for (j = 0; j < elements[i].nne; j++)
      fprintf (out," %ld",elements[i].nodes[j]+1);
    fprintf (out," %ld",elements[i].prop);
    if (elements[i].propedg)
    {
      fprintf(out, "  ");
      for (j=0; j < elements[i].ned; j++)
        fprintf (out," %ld",elements[i].propedg[j]);
    }
    if (elements[i].propsurf)
    {
      fprintf(out, "  ");
      for (j=0; j < elements[i].nsurf; j++)
        fprintf (out," %ld",elements[i].propsurf[j]);
    }
    fprintf(out, "\n");
  }    
  if (edges)
    edges->print(out, NULL);
  if (surfaces)
    surfaces->print(out, NULL);
}



/**
  The function prints nodes to the opened text file with the shift indeces.

  @param out - pointer to the opened text file
  @param shift - shift for nodal indeces

  @return The function does not return anything.
 
  9.11.2010 JB
*/
void siftop::shift_print_nodes (FILE *out,long shift)
{
  long i,j;
 for (i = 0; i < nn; i++){
   fprintf (out,"%ld %15.10le %15.10le %15.10le %ld",i+1+shift, nodes[i].x, nodes[i].y, nodes[i].z, nodes[i].nprop);
   for(j=0; j<nodes[i].nprop; j++)
     {
       fprintf(out, " %d %ld", nodes[i].entid[j], nodes[i].prop[j]);
     }
   fprintf(out, "\n");
 }
}


/**
  The function prints elements to the opened text file with the shift indeces.

  @param out - pointer to the opened text file
  @param shift - shift for element indeces

  @return The function does not return anything.
 
  9.11.2010 JB
*/
void siftop::shift_print_elements (FILE *out,long shiftnn,long shiftne){
  
  long i,j;
  for (i = 0; i < ne; i++){
    fprintf (out,"%ld %d",i+1+shiftne, elements[i].type);
    for (j = 0; j < elements[i].nne; j++)
      fprintf (out," %ld",elements[i].nodes[j]+1+shiftnn);
    fprintf (out," %ld",elements[i].prop);
    if (elements[i].propedg)
    {
      fprintf(out, "  ");
      for (j=0; j < elements[i].ned; j++)
        fprintf (out," %ld",elements[i].propedg[j]);
    }
    if (elements[i].propsurf)
    {
      fprintf(out, "  ");
      for (j=0; j < elements[i].nsurf; j++)
        fprintf (out," %ld",elements[i].propsurf[j]);
    }
    fprintf(out, "\n");
  }    
}



/**
  Method gets set of nodes with property prop of entity ent
  Node numbers of found nodes are stored to array setnodes

  @param prop  - searched property id 
  @param ent   - type of searched entity, in case of gsurface, 
                 the esurface, epatch and eshell are assumed to be equal to gsurface
  @param setnodes - allocated array for node numbers with given property of given entity
  
  @return The function returns number of nodes with checked property. 

  Created 06.2009 by Tomas Koudelka koudelka@cml.fsv.cvut.cz                
  Modified 08.2011 by Tomas Koudelka
*/
long siftop::get_propent_nodes(long prop, gentity ent, long *setnodes)
{
  long i, j, ret=0;
  for (i=0; i<nn; i++)
  {
    setnodes[i] = -1;
    if (nodes[i].searchprop(prop, ent, i, NULL, NULL))
    {
      setnodes[i] = i;
      ret++;
    }
  }
  if ((ent == gcurve) && (edges))
  {
    // search additional edges specified in the MeshEditor
    for(i=0; i<edges->nprop; i++)
    {
      if (edges->proplist[i] == prop)
      {
        for(j=0; j<edges->prop_list_nod_num[i]; j++)
        {
          setnodes[edges->prop_list_nod[i][j]] = edges->prop_list_nod[i][j];
          ret++;
        }
      }
    }
  }

  if ((ent == gsurface) && (surfaces))
  {
    // search additional surfacs specified in the MeshEditor
    for(i=0; i<surfaces->nprop; i++)
    {
      if (surfaces->proplist[i] == prop)
      {
        for(j=0; j<surfaces->prop_list_nod_num[i]; j++)
        {
          setnodes[surfaces->prop_list_nod[i][j]] = surfaces->prop_list_nod[i][j];
          ret++;
        }
      }
    }
  }
  return ret;
}



/**
  Method gets ndof of nodes with property prop of entity ent
  All nodes with given property of entity are searched and checked that
  they have same ndof. Node numbers of found nodes are stored to array setnodes

  @param prop  - searched property id 
  @param ent   - type of searched entity, in case of gsurface, 
                 the esurface, epatch and eshell are assumed to be equal to gsurface
  @param ndofn - array of ndofs for all nodes
  @param setnodes - allocated array for node numbers with given property of given entity
  
  @retval >0 - ndof of nodes with given property of given entity
  @retval 0  - in case that the nodes with given property of given entity have the different ndof
  @retval <0 - in case that no nodes with the given property were found

  Created 04.2008 by Tomas Koudelka koudelka@cml.fsv.cvut.cz                
*/
 long siftop::get_ndofn(long prop, gentity ent, long *ndofn, long *setnodes)
{
  long i, j, id;
  long ndof = -1;  
  for (i=0; i<nn; i++)
  {
    setnodes[i] = -1;
    if (nodes[i].searchprop(prop, ent, i, NULL, NULL))
    {
      setnodes[i] = i;
      if (ndof == -1)
        ndof = ndofn[i];
      else
      {
        if (ndof != ndofn[i])
          return 0;
      }
    }
  }


  if ((ent == gcurve) && (edges))
  {
    // search additional edges specified in the MeshEditor
    for(i=0; i<edges->nprop; i++)
    {
      if (edges->proplist[i] == prop)
      {
        for(j=0; j<edges->prop_list_nod_num[i]; j++)
        {
          id = edges->prop_list_nod[i][j];
          setnodes[id] = id;
          if (ndof == -1)
            ndof = ndofn[id];
          else
          {
            if (ndof != ndofn[id])
              return 0;
          }
        }
      }
    }
  }

  if ((ent == gsurface) && (surfaces))
  {
    // search additional surfacs specified in the MeshEditor
    for(i=0; i<surfaces->nprop; i++)
    {
      if (surfaces->proplist[i] == prop)
      {
        for(j=0; j<surfaces->prop_list_nod_num[i]; j++)
        {
          id = surfaces->prop_list_nod[i][j];
          setnodes[id] = id;
          if (ndof == -1)
            ndof = ndofn[id];
          else
          {
            if (ndof != ndofn[id])
              return 0;
          }
        }
      }
    }
  }
  return ndof;
}




/**
  The function exports sets of used elements to the file given by parameter out in GiD format. 

  @param out  - pointer to the opened text file where the output will be produced
  @param idn1 - id of the first node for GiD mesh (default should be idn1 = 1 -> nodes are numbered from 1)
  @param ide1 - id of the first element for GiD mesh (default should be ide1 = 1 -> elements are numbered from 1)

  @return The function does not return anything.

  Created 2004 by Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
void siftop::export_gid_mesh(FILE *out, long idn1, long ide1)
{
  long i, print_header, print_coord = 1;

  print_header = 1;
  for (i=0; i < ne; i++)
  {
    if (elements[i].type == isolinear1d)
    {
      if (print_header)
      {
        fprintf(out, "MESH \"%ld-beams2\" dimension 3  Elemtype Linear Nnode 2\n", ide1);
        print_header = 0;
        if (print_coord)
	{
          write_gid_nodes(out, idn1);
          print_coord = 0;
        }
        fprintf(out, "Elements\n");
      }
      write_gid_element(out, i, idn1, ide1);
    }
  }
  if (print_header == 0)
    fprintf(out, "end Elements\n");

  print_header = 1;
  for (i=0; i < ne; i++)
  {
    if (elements[i].type == isoquadratic1d)
    {
      if (print_header)
      {
        fprintf(out, "MESH \"%ld-beams3\" dimension 3  Elemtype Linear Nnode 3\n", ide1);
        print_header = 0;
        if (print_coord)
	{
          write_gid_nodes(out, idn1);
          print_coord = 0;
        }
        fprintf(out, "Elements\n");
      }
      write_gid_element(out, i, idn1, ide1);
    }
  }
  if (print_header == 0)
    fprintf(out, "end Elements\n");

  print_header = 1;
  for (i=0; i < ne; i++)
  {
    if (elements[i].type == trianglelinear)
    {
      if (print_header)
      {
        fprintf(out, "MESH \"%ld-trias3\" dimension 3  Elemtype Triangle Nnode 3\n", ide1);
        print_header = 0;
        if (print_coord)
	{
          write_gid_nodes(out, idn1);
          print_coord = 0;
        }
        fprintf(out, "Elements\n");
      }
      write_gid_element(out, i, idn1, ide1);
    }
  }
  if (print_header == 0)
    fprintf(out, "end Elements\n");

  print_header = 1;
  for (i=0; i < ne; i++)
  {
    if (elements[i].type == trianglequadratic)
    {
      if (print_header)
      {
        fprintf(out, "MESH \"%ld-trias6\" dimension 3  Elemtype Triangle Nnode 6\n", ide1);
        print_header = 0;
        if (print_coord)
	{
          write_gid_nodes(out, idn1);
          print_coord = 0;
        }
        fprintf(out, "Elements\n");
      }
      write_gid_element(out, i, idn1, ide1);
    }
  }
  if (print_header == 0)
    fprintf(out, "end Elements\n");

  print_header = 1;
  for (i=0; i < ne; i++)
  {
    if (elements[i].type == isolinear2d)
    {
      if (print_header)
      {
        fprintf(out, "MESH \"%ld-Quads4\" dimension 3  Elemtype Quadrilateral Nnode 4\n", ide1);
        print_header = 0;
        if (print_coord)
	{
          write_gid_nodes(out, idn1);
          print_coord = 0;
        }
        fprintf(out, "Elements\n");
      }
      write_gid_element(out, i, idn1, ide1);
    }
  }
  if (print_header == 0)
    fprintf(out, "end Elements\n");

  print_header = 1;
  for (i=0; i < ne; i++)
  {
    if (elements[i].type == isoquadratic2d)
    {
      if (print_header)
      {
        fprintf(out, "MESH \"%ld-Quads8\" dimension 3  Elemtype Quadrilateral Nnode 8\n", ide1);
        print_header = 0;
        if (print_coord)
	{
          write_gid_nodes(out, idn1);
          print_coord = 0;
        }
        fprintf(out, "Elements\n");
      }
      write_gid_element(out, i, idn1, ide1);
    }
  }
  if (print_header == 0)
    fprintf(out, "end Elements\n");

  print_header = 1;
  for (i=0; i < ne; i++)
  {
    if (elements[i].type == tetrahedronlinear)
    {
      if (print_header)
      {
        fprintf(out, "MESH \"%ld-Tetras4\" dimension 3  Elemtype Tetrahedra Nnode 4\n", ide1);
        print_header = 0;
        if (print_coord)
        {
          write_gid_nodes(out, idn1);
          print_coord = 0;
        }
        fprintf(out, "Elements\n");
      }
      write_gid_element(out, i, idn1, ide1);
    }
  }
  if (print_header == 0)
    fprintf(out, "end Elements\n");

  print_header = 1;
  for (i=0; i < ne; i++)
  {
    if (elements[i].type == tetrahedronquadratic)
    {
      if (print_header)
      {
        fprintf(out, "MESH \"%ld-Tetras10\" dimension 3  Elemtype Tetrahedra Nnode 10\n", ide1);
        print_header = 0;
        if (print_coord)
	{
          write_gid_nodes(out, idn1);
          print_coord = 0;
        }
        fprintf(out, "Elements\n");
      }
      write_gid_element(out, i, idn1, ide1);
    }
  }
  if (print_header == 0)
    fprintf(out, "end Elements\n");

  print_header = 1;
  for (i=0; i < ne; i++)
  {
    if (elements[i].type == isolinear3d)
    {
      if (print_header)
      {
        fprintf(out, "MESH \"%ld-Brick8\" dimension 3  Elemtype Hexahedra Nnode 8\n", ide1);
        print_header = 0;
        if (print_coord)
	{
          write_gid_nodes(out, idn1);
          print_coord = 0;
        }
        fprintf(out, "Elements\n");
      }
      write_gid_element(out, i, idn1, ide1);
    }
  }
  if (print_header == 0)
    fprintf(out, "end Elements\n");

  print_header = 1;
  for (i=0; i < ne; i++)
  {
    if (elements[i].type == isoquadratic3d)
    {
      if (print_header)
      {
        fprintf(out, "MESH \"%ld-Brick20\" dimension 3  Elemtype Hexahedra Nnode 20\n", ide1);
        print_header = 0;
        if (print_coord)
	{
          write_gid_nodes(out, idn1);
          print_coord = 0;
        }
        fprintf(out, "Elements\n");
      }
      write_gid_element(out, i, idn1, ide1);
    }
  }
  if (print_header == 0)
    fprintf(out, "end Elements\n");
}



/**
  The function exports nodes to the file given by parameter out in GiD format. 

  @param out  - pointer to the opened text file where the output will be produced
  @param idn1 - id of the first node for GiD mesh (default should be idn1 = 1 -> nodes are numbered from 1)

  @return The function does not return anything.

  Created 2010 by Tomas Koudelka, koudelka@cml.fsv.cvut.cz
 */
void siftop::write_gid_nodes(FILE *out, long idn1)
{
  long i;
  fprintf(out, "Coordinates\n");
  for (i=0; i<nn; i++)
    fprintf(out, "%ld %e %e %e\n", i+idn1, nodes[i].x, nodes[i].y, nodes[i].z);
  fprintf(out, "end Coordinates\n");
}



/**
  The function exports element to the file given by parameter out in GiD format. 

  @param out  - pointer to the opened text file where the output will be produced
  @param i    - element id
  @param idn1 - id of the first node for GiD mesh (default should be idn1 = 1 -> nodes are numbered from 1)
  @param ide1 - id of the first element for GiD mesh (default should be ide1 = 1 -> elements are numbered from 1)

  @return The function does not return anything.

  Created 2010 by Tomas Koudelka, koudelka@cml.fsv.cvut.cz
 */
void siftop::write_gid_element(FILE *out, long i, long idn1, long ide1)
{
  long j;
  fprintf(out, "%ld ", i+ide1);
  for (j=0; j<elements[i].nne; j++)
    fprintf(out, "%ld ", elements[i].nodes[j]+idn1);
  fprintf(out, "%ld\n", elements[i].prop);
}



/**
  The function reorders mesh form T3d into sifel format.

  @return The function does not return anything.

  Created by JB
*/
void siftop::t3d2sifelformat()
{
  long i,j;
  long *help;
  help = new long[20];
  
  for(i = 0; i < ne; i++){
    switch(elements[i].type){
    case tetrahedronlinear:{
      for(j = 0; j < 4; j++){
	help[j] = elements[i].nodes[j];
      }
      elements[i].nodes[2] = help[3];
      elements[i].nodes[3] = help[2];
	      
      for(j = 0; j < 4; j++){
	help[j] = elements[i].propsurf[j];
      }
      elements[i].propsurf[0]= help[2];
      elements[i].propsurf[1]= help[3];
      elements[i].propsurf[2]= help[0];
      elements[i].propsurf[3]= help[1];
      
      break;
    }
    case tetrahedronquadratic:{
      
      for(j = 0; j < 10; j++){
	help[j] = elements[i].nodes[j];
      }
      
      elements[i].nodes[2] = help[3];
      elements[i].nodes[3] = help[2];
      elements[i].nodes[5] = help[8];
      elements[i].nodes[6] = help[7];
      elements[i].nodes[7] = help[6];
      elements[i].nodes[8] = help[5];
      
      for(j = 0; j < 4; j++){
	help[j] = elements[i].propsurf[j];
      }
      elements[i].propsurf[0]= help[2];
      elements[i].propsurf[1]= help[3];
      elements[i].propsurf[2]= help[0];
      elements[i].propsurf[3]= help[1];
      
      break;
    }
    case isolinear3d:{
      for(j = 0; j < 8; j++){
	help[j] = elements[i].nodes[j];
      }
      
      elements[i].nodes[0] = help[3];
      elements[i].nodes[1] = help[0];
      elements[i].nodes[2] = help[4];
      elements[i].nodes[3] = help[7];
      elements[i].nodes[4] = help[2];
      elements[i].nodes[5] = help[1];
      elements[i].nodes[6] = help[5];
      elements[i].nodes[7] = help[6];
      
      
      for(j = 0; j < 6; j++){
	help[j] = elements[i].propsurf[j];
      }

      elements[i].propsurf[0] = help[0];
      elements[i].propsurf[1] = help[3];
      elements[i].propsurf[2] = help[1];
      elements[i].propsurf[3] = help[5];
      elements[i].propsurf[4] = help[2];
      elements[i].propsurf[5] = help[4];
      
      
      break;
    }
    case isoquadratic3d:{
      for(j = 0; j < 20; j++){
	help[j] = elements[i].nodes[j];
      }
      
      elements[i].nodes[0] = help[3];
      elements[i].nodes[1] = help[0];
      elements[i].nodes[2] = help[4];
      elements[i].nodes[3] = help[7];
      elements[i].nodes[4] = help[2];
      elements[i].nodes[5] = help[1];
      elements[i].nodes[6] = help[5];
      elements[i].nodes[7] = help[6];
      elements[i].nodes[8] = help[11];
      elements[i].nodes[9] = help[16];
      elements[i].nodes[10] = help[15];
      elements[i].nodes[11] = help[19];
      elements[i].nodes[12] = help[10];
      elements[i].nodes[13] = help[8];
      elements[i].nodes[14] = help[12];
      elements[i].nodes[15] = help[14];
      elements[i].nodes[16] = help[9];
      elements[i].nodes[17] = help[17];
      elements[i].nodes[18] = help[13];
      elements[i].nodes[19] = help[18];
      
      for(j = 0; j < 6; j++){
	help[j] = elements[i].propsurf[j];
      }
      elements[i].propsurf[0] = help[0];
      elements[i].propsurf[1] = help[3];
      elements[i].propsurf[2] = help[1];
      elements[i].propsurf[3] = help[5];
      elements[i].propsurf[4] = help[2];
      elements[i].propsurf[5] = help[4];
      
      break;
    }
    default:{
      break;
    }
    }
  }
  delete []help;
  
}
